package clustering.stc;

import tools.StopwordRemover;

import java.util.Map.Entry;
import java.util.*;

import searching.Settings;
import clustering.suffixtree.SuffixTree;

public class SuffixTreeClusterer {

    SuffixTree suffixTree;
    final Map wordBoosts; //Maps word to boost factor by ODP
    final Set boostWords; //Contains all queries

    public SuffixTreeClusterer(SuffixTree tree, Map wordBoosts, Set boostWords) {
        this.suffixTree = tree;
        this.wordBoosts = wordBoosts;
        this.boostWords = boostWords;
    }

    public SuffixTreeClusterer(ArrayList snippets, ArrayList docIDs, Map wordBoosts, Set boostWords) {
        this.wordBoosts = wordBoosts;
        this.boostWords = boostWords;
        this.suffixTree = new SuffixTree();
        for (int i = 0; i < snippets.size(); i++)
            suffixTree.addSnippet((String) snippets.get(i), (Integer) docIDs.get(i));
    }

    public Cluster[] clusterDocs() {
        ArrayList baseClusters = suffixTree.getBaseClusters(); //ArrayList of BaseCluster
        float[] baseClusterScores = calculateBaseClusterScore(baseClusters);

        //Remove zeroScoring BaseClusters and add scores to the rest
        for (int i = baseClusterScores.length - 1; i >= 0; i--) {
            if (baseClusterScores[i] == 0.0f) {
                baseClusters.remove(i);
            } else {
                ((BaseCluster) baseClusters.get(i)).score = baseClusterScores[i];
            }
        }

        //Calculate similerites and generate clusters
        int[][] clusterGraph = calculateSimilarities(baseClusters);
        Cluster[] clusters = getConnectedComponents(clusterGraph, baseClusters);

        //Update the list of IDs contained in each cluster and label them
        for (int i = 0; i < clusters.length; i++) {
            Cluster cluster = clusters[i];
            Set IDs = new HashSet();

            //Gather all docIDs from all clusters in this Set and add to IDs
            BaseCluster[] BCs = (BaseCluster[]) cluster.baseClusters.toArray(new BaseCluster[cluster.baseClusters.size()]);
            for (int j = 0; j < BCs.length; j++) {
                BaseCluster bc = BCs[j];
                Set docIDs = bc.docIDs;
                IDs.addAll(docIDs);
            }
            cluster.docIDs = IDs;

            //Calculate coverage:
            for (int j = 0; j < BCs.length; j++) {
                BaseCluster bc = BCs[j];
                bc.coverage = (float) bc.docIDs.size() / (float) cluster.docIDs.size();
            }

            //Label the cluster
            cluster.label = labelCluster(BCs);
        }

        Arrays.sort(clusters);
        return clusters;
    }

    private String labelCluster(BaseCluster[] BCs) {

        //Label cluster with highestScoring baseCluster
        if (BCs.length > 1) //Only sort if it contains multiple clusters
            Arrays.sort(BCs);

        return getLabelForBaseCluster(BCs[0]);
    }

    public String getLabelForBaseCluster(BaseCluster bc) {
        //Construct original label
        String label = suffixTree.getOriginalLabel(bc);

        //CamelCase the label
        char[] chars = label.toCharArray();
        chars[0] = Character.toUpperCase(chars[0]);
        for (int pos = 0; pos < label.length(); pos++) {
            pos = label.indexOf(' ', pos);
            if (pos == -1)
                break;
            chars[pos + 1] = Character.toUpperCase(chars[pos + 1]);
        }

        return new String(chars);
    }

    private Cluster[] getConnectedComponents(int[][] clusterGraph, ArrayList baseClusters) {

        int nofBaseClusters = baseClusters.size();
        HashMap sets = new HashMap(nofBaseClusters);

        //MakeSets
        for (int i = 0; i < nofBaseClusters; i++) {
            Set li = new HashSet(); //The list is the representative
            BaseCluster cluster = (BaseCluster) baseClusters.get(i);
            li.add(cluster);
            sets.put(cluster, li); //All objects are mapped to the arrayList
        }

        //AddEdges
        for (int i = 0; i < nofBaseClusters; i++) {
            for (int j = i + 1; j < nofBaseClusters; j++) {
                if (clusterGraph[i][j] == 1) {
                    //Add edge from i to j. a and b represent the Sets of Clusters i and j
                    Set a = (Set) sets.get(baseClusters.get(i));
                    Set b = (Set) sets.get(baseClusters.get(j));
                    if (a != b) {
                        //Join a and b
                        sets.remove(baseClusters.get(j)); //Remove set J
                        a.addAll(b); //Add all baseClusters of J to I

                        //Map all BaseClusters of b to a
                        for (Iterator iterator = b.iterator(); iterator.hasNext();) { //Map each object in J to I
                            sets.put(iterator.next(), a); //denne lager ny mapping, ikke til samme object..
                        }
                    }
                }
            }
        }

        //Get uniqueSets (some map to the same Set (sets.values is not a Set)
        Set uniqueSets = new HashSet();
        for (Iterator iterator = sets.values().iterator(); iterator.hasNext();) {
            uniqueSets.add((Set) iterator.next());
        }

        //Generate clusterarray:
        Cluster[] clusters = new Cluster[uniqueSets.size()];
        int index = 0;
        for (Iterator iterator = uniqueSets.iterator(); iterator.hasNext();) {
            clusters[index++] = new Cluster(null, (Set) iterator.next());
        }

        return clusters;
    }

    //Returns 0-1 matrix where 1 indicates that sets[i] and [j] are connected
    private int[][] calculateSimilarities(ArrayList baseClusters) {
        int nofClusters = baseClusters.size();
        int[][] similarities = new int[nofClusters][nofClusters];
        //TODO: fortere  putte getConnectedComponents rett inn her?
        for (int i = 0; i < nofClusters; i++) {
            for (int j = i + 1; j < nofClusters; j++) {
                similarities[i][j] = similarity(((BaseCluster) baseClusters.get(i)).docIDs, ((BaseCluster) baseClusters.get(j)).docIDs);
            }
        }
        return similarities;
    }

    private int similarity(Set a, Set b) {

        //Get intersectionSize:
        int nof = 0;
        int aSize = a.size();
        int bSize = b.size();
        for (Iterator iterator = a.iterator(); iterator.hasNext();) {
            if (b.contains(iterator.next()))
                nof++;
        }

        //Calculate similarity
        float first = (float) nof / (float) aSize;
        float second = (float) nof / (float) bSize;
        return (first > Settings.MIN_BASE_CLUSTER_OVERLAP && second > Settings.MIN_BASE_CLUSTER_OVERLAP) ? 1 : 0;
    }

    //Calculates similarity for each baseCluster
    private float[] calculateBaseClusterScore(ArrayList baseClusters) { //BaseCluster[]
        Map wordToDocIDSet = suffixTree.getWordToDocIDSet();
        int nofDocs = suffixTree.nofDocs();
        int maxLimit = Math.round(Settings.MAX_PERCENTAGE_OF_DOCS_WORD_APPEARS_IN * nofDocs); //Occurring more often than this will give similarity = 0
        int minLimit = Settings.MIN_NOF_DOCS_WORD_APPEARS_IN;

        //Find zeroscoring words;
        Set zeroScoringWords = new HashSet();
        for (Iterator iterator = wordToDocIDSet.entrySet().iterator(); iterator.hasNext();) {
            Entry entry = (Entry) iterator.next();
            String word = (String) entry.getKey();
            int docOccs = ((Set) entry.getValue()).size();
            if (docOccs <= minLimit || docOccs >= maxLimit)
                zeroScoringWords.add(word);
        }

        StopwordRemover sr = StopwordRemover.getInstance();
        float[] scores = new float[baseClusters.size()];

        for (int i = 0; i < baseClusters.size(); i++) {
            BaseCluster cluster = (BaseCluster) baseClusters.get(i);
            String label = cluster.label;
            String[] tokens = label.split(" ");
            int len = 0;
            for (int j = 0; j < tokens.length; j++) {
                String s = tokens[j];
                if (!sr.isStopword(s) && !zeroScoringWords.contains(s))
                    len++;
            }

            if (len == 0) {
                //Last part of s(B) will be 0
                scores[i] = 0;
                continue;
            }

            //Caluculate f(len)
            float lengthScore;
            if (len == 1)
                lengthScore = 0.5f;
            else if (len <= 6)
                lengthScore = len;
            else
                lengthScore = 7;

            //Calculate |B|
            int nofDocsInCluster = cluster.docIDs.size();
            scores[i] = nofDocsInCluster * lengthScore; //s(B) = |B| * f(|P|) = nofDocsInCluster(B) * f(nofEffectivePhrases)

            /*cluster.b_length = nofDocsInCluster;
            cluster.f_of_p_length = lengthScore;

            //Calculate average boost factor
            float ODPboost = 0;
            for (int j = 0; j < tokens.length; j++) {
                String token = tokens[j];
                Float wordBoost = (Float) wordBoosts.get(token);
                if (wordBoost == null)
                //set boost = 1
                    ODPboost += 1;
                else
                    ODPboost += wordBoost.floatValue();
            }
            ODPboost /= (float) tokens.length;
            cluster.odpboost = ODPboost;

            //Calculate average boost factor
            float LOGboost = 0;
            for (int j = 0; j < tokens.length; j++) {
                String token = tokens[j];
                if (boostWords.contains(token))
                    LOGboost += 1;
            }
            //LOGboost /= (float) tokens.length;
            cluster.queryboost = LOGboost;*/

            /*if (Settings.BOOST_ODP_LABELS) {
                //Calculate average boost factor
                float boost = 0;
                for (int j = 0; j < tokens.length; j++) {
                    String token = tokens[j];
                    Float wordBoost = (Float) wordBoosts.get(token);
                    if (wordBoost == null)
                        //set boost = 1
                        boost += 1;
                    else
                        boost += wordBoost.floatValue();
                }
                boost /= (float) tokens.length;
                boost *= 2;

                //System.out.print("Boosted '" + cluster.label + "' fra " + scores[i]);
                scores[i] *= boost;
                //System.out.println(" til " + scores[i] + " med " + boost);
            }*/
        }
        return scores;
    }
}
