package searching.engine;

import searching.Settings;
import searching.dataobjects.VocabularyEntry;
import searching.dataobjects.DocumentManager;
import searching.dataobjects.Result;

import java.io.*;
import java.util.*;

import tools.PorterStemmer;

public class Index {

    private RandomAccessFile postingsFile;
    private int nofDocs;
    private VocabularyEntry[] vocabulary;
    private DocumentManager docManager;
    private PorterStemmer stemmer;

    public Index(String indexFolder) {
        String postingsFilename = indexFolder + "/postings.dat";
        String vocabularyFilename = indexFolder + "/vocabulary.dat";
        String docInfoFilename = indexFolder + "/docinfo.dat";
        readVocabularyFromFile(vocabularyFilename);
        docManager = new DocumentManager(docInfoFilename);
        nofDocs = docManager.getNofDocs();
        stemmer = PorterStemmer.getInstance();
        try {
            postingsFile = new RandomAccessFile(postingsFilename, "rw");
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

   public Result[] query(String query) {

        String[] querytokens = tokenizeQuery(query);
        StringBuffer sb = new StringBuffer("Searching for '" + querytokens[0]);
        for (int i = 1; i < querytokens.length; i++)
            sb.append(" " + querytokens[i]);
        sb.append("':");
        //System.out.println(sb.toString());

        ArrayList similarities = calculateSimilarities(querytokens);
        ArrayList pairs = new ArrayList(100);
        float treshold = 0.001f;
        for (int i = 0; i < similarities.size(); i++) {
            int docID = ((Integer) similarities.get(i++)).intValue();
            float similarity = ((Float) similarities.get(i++)).floatValue();
            int filepointer = ((Integer) similarities.get(i)).intValue();
            if (similarity > treshold) {
                pairs.add(new Result(docID, filepointer, similarity));
            }
        }
        Result[] results = (Result[]) pairs.toArray(new Result[pairs.size()]);
        Arrays.sort(results);
        return results;
    }

    private ArrayList calculateSimilarities(String[] tokens) {
        ArrayList docIDAndSimilarities = new ArrayList(); //docID (Integer) and similarity (Float) for each hit
        Object[] queryPosFreqsAndMaxFreq = getQueryVectorWithPositions(tokens);
        int[] queryPositions = (int[]) queryPosFreqsAndMaxFreq[0]; //dictionaryEntryPosition for each queryToken
        int[] queryFreqs = (int[]) queryPosFreqsAndMaxFreq[1]; //frequency for each queryToken
        int queryMaxTermFreq = ((Integer) queryPosFreqsAndMaxFreq[2]).intValue();
        int queryLength = queryPositions.length;
        float[] queryWeights = new float[queryLength]; //Holds weights for each token in the query
        int[] nofDocsContainingTerm = new int[queryLength]; //Holds nofDocsContainingTerm for each of the tokens
        HashMap docIDToWeights = new HashMap(); //Maps docID to float[] of weights

        //For each token in the query, calculate the tfIdfScore of the documents in the resultset
        for (int queryTokenIndex = 0; queryTokenIndex < queryPositions.length; queryTokenIndex++) {
            VocabularyEntry entry = vocabulary[queryPositions[queryTokenIndex]];
            float idf = (float) Math.log((float) nofDocs / entry.nofDocs);
            nofDocsContainingTerm[queryTokenIndex] = entry.nofDocs;
            try {
                //Calculate weight for each document for this term
                postingsFile.seek(entry.filePointer);
                for (int currentDocumentIndex = 0; currentDocumentIndex < entry.nofDocs; currentDocumentIndex++) {
                    int docID = postingsFile.readInt();
                    int freq = postingsFile.readShort();
                    int filepointer = postingsFile.readInt();
                    float tf = (float) freq / docManager.getHighestFreq(docID);
                    float tfIdf = tf * idf;
                    Integer docIndex = new Integer(docID);
                    float[] weights;
                    if (!docIDToWeights.containsKey(docIndex)) {
                        weights = new float[queryLength + 1]; //Litt skitten, siste holder filepointer
                        weights[queryLength] = Float.MAX_VALUE; //Gr bra fordi ordet alltid vil finnes...
                        docIDToWeights.put(docIndex, weights);
                    } else {
                        weights = (float[]) docIDToWeights.get(docIndex);
                    }
                    weights[queryTokenIndex] = tfIdf; //Store tfIdf at this tokens position in the queryVector
                    //Store smallest filepointer to queryword in this document
                    if (weights[queryLength] > filepointer)
                        weights[queryLength] = filepointer;
                }

                //Calculate queryWeight for this term (Salton & Buckley suggestion)
                float tf = (float) (0.5 + (0.5 * queryFreqs[queryTokenIndex]) / (float) queryMaxTermFreq);
                queryWeights[queryTokenIndex] = tf * idf;

            } catch (IOException e) {
                e.printStackTrace();
            }
        }

        //Calculate |query|
        float queryVectorLength = 0;
        for (int i = 0; i < queryLength; i++) {
            queryVectorLength += queryWeights[i] * queryWeights[i];
        }
        queryVectorLength = (float) Math.sqrt(queryVectorLength);

        //Calculate similarities for each of the docs in the resultset (docIDs) (Wj x Wq) / (|j| * |q|)
        Set docIDs = docIDToWeights.keySet();
        for (Iterator iterator = docIDs.iterator(); iterator.hasNext();) {
            Integer docID = (Integer) iterator.next();
            float[] weights = (float[]) docIDToWeights.get(docID);
            float nominator = 0;
            for (int i = 0; i < queryLength; i++) {
                nominator += weights[i] * queryWeights[i];
            }
            float similarity = nominator / (docManager.getVectorLength(docID.intValue()) * queryVectorLength);
            docIDAndSimilarities.add(docID);
            docIDAndSimilarities.add(new Float(similarity));
            docIDAndSimilarities.add(new Integer((int) weights[queryLength]));
        }

        return docIDAndSimilarities;
    }

    //Return {int[] with pos, int[] with weights, int maxFreqOfTermInQuery} of terms occurring in the engine
    private Object[] getQueryVectorWithPositions(String[] tokens) {
        HashMap posToFreq = new HashMap(); //maps pos as Integer to Freq as Integer
        for (int i = 0; i < tokens.length; i++) {
            int pos = getDictionaryEntryPosition(tokens[i]);
            if (pos != -1) {
                Integer posInt = new Integer(pos);
                if (posToFreq.containsKey(posInt))
                    posToFreq.put(posInt, new Integer(((Integer) posToFreq.get(posInt)).intValue() + 1)); //Inc freq
                else
                    posToFreq.put(posInt, new Integer(1)); //Add new pos with freq 1
            }
        }

        int[] pos = new int[posToFreq.size()];
        int[] freq = new int[posToFreq.size()];
        Set positions = posToFreq.keySet();
        int counter = 0;
        int maxFreqOfTermInQuery = 0; //Needed for caculating querytermweights
        for (Iterator iterator = positions.iterator(); iterator.hasNext();) {
            Integer position = (Integer) iterator.next();
            pos[counter] = position.intValue();
            int frequency = ((Integer) posToFreq.get(position)).intValue();
            if (frequency > maxFreqOfTermInQuery)
                maxFreqOfTermInQuery = frequency;
            freq[counter] = frequency;
            counter++;
        }
        return new Object[]{pos, freq, new Integer(maxFreqOfTermInQuery)};
    }

    // Gets words pos in vocabulary or -1 if not found (binarysearch)
    private int getDictionaryEntryPosition(String word) {
        int low = 0;
        int high = vocabulary.length - 1;

        while (low <= high) {
            int mid = (low + high) >> 1;
            VocabularyEntry de = vocabulary[mid];
            int cmp = de.word.compareTo(word);
            if (cmp < 0)
                low = mid + 1;
            else if (cmp > 0)
                high = mid - 1;
            else
                return mid;//key found
        }
        return -1;//key not found.
    }

    private String[] tokenizeQuery(String query) {
        StringBuffer sb = new StringBuffer();
        char[] chars = query.toCharArray();
        char temp;
        ArrayList tokens = new ArrayList(100);

        for (int i = 0; i < chars.length; i++) {
            temp = chars[i];
            if (temp == ' ' || temp == '.' || temp == ',' || temp == '\n' || temp == '\t' || temp == ':' || temp == '\"' || temp == '\'' || temp == '-') {
                //Found specialchar. If len>0 make word, and reset stringbuffer.
                if (sb.length() > 0) {
                    String word = sb.toString().toLowerCase();
                    if (Settings.USE_STEMMING)
                        word = stemmer.getStemmedWord(word);
                    tokens.add(word);
                    sb = new StringBuffer();
                }
            } else {
                //Found part of a word
                sb.append(temp);
            }
        }

        //Add last token (potentially)
        if (sb.length() > 0) {
            String word = sb.toString().toLowerCase();
            if (Settings.USE_STEMMING)
                word = stemmer.getStemmedWord(word);
            tokens.add(word);
        }

        return (String[]) tokens.toArray(new String[tokens.size()]);
    }

    public void readVocabularyFromFile(String vocabularyFilename) {
        long start = System.currentTimeMillis();
        ObjectInputStream vocabularyFile;
        try {
            vocabularyFile = new ObjectInputStream(new FileInputStream(vocabularyFilename));
            int dictLength = vocabularyFile.readInt();
            vocabulary = new VocabularyEntry[dictLength];
            for (int i = 0; i < vocabulary.length; i++) {
                //Raskere med en readUTF?
                vocabulary[i] = new VocabularyEntry((String) vocabularyFile.readObject(), vocabularyFile.readInt(), vocabularyFile.readLong());
            }
            //vocabulary = (VocabularyEntry[]) vocabularyFile.readObject();
            vocabularyFile.close();
        } catch (IOException e) {
            e.printStackTrace();
        } catch (ClassNotFoundException e) {
            e.printStackTrace();
        }
        //System.out.println("Read vocabulary in " + (System.currentTimeMillis() - start) + "ms.");
    }

    public DocumentManager getDocumentManager() {
        return docManager;
    }
}
