package searching.engine;

import tools.StopwordRemover;
import tools.StopWatch;
import tools.PorterStemmer;

import java.io.*;
import java.util.*;

import searching.Settings;
import searching.dataobjects.VocabularyEntry;
import searching.dataobjects.Document;

public class IndexWriter {

    private HashMap termsToFreqList = new HashMap(); //Maps words as string to ArrayList of Frequencys
    private VocabularyEntry[] vocabulary;
    private Document[] documents;
    private short[] highestFreq; //Stores the frequency for most frequent word in document i
    private float[] vectorLengths; //Sqrt of sum of squared weights (see cosine similarity)
    private int nofTermsInIndex;
    private final int maxTokensPerDocument;
    private int nofDocs = 0;

    private static StopwordRemover stopwordRemover;
    private static PorterStemmer stemmer;
    private String documentFolder;

    private final String postingsFilename;
    private final String vocabularyFilename;
    private final String docInfoFilename;
    private RandomAccessFile postingsFile;

    private int nofFilesLeft;
    private int interval;

    public static void main(String[] args) {
        IndexWriter writer = new IndexWriter("indices/queries", 200);
        writer.buildIndexFromFolder("collections/queriesStripped", 3880);
    }

    public IndexWriter(String indexFolder, int maxTokensPerDocument) {
        File folder = new File(indexFolder);
        folder.mkdirs();
        postingsFilename = indexFolder + "/postings.dat";
        vocabularyFilename = indexFolder + "/vocabulary.dat";
        docInfoFilename = indexFolder + "/docinfo.dat";
        this.maxTokensPerDocument = maxTokensPerDocument;
        stopwordRemover = StopwordRemover.getInstance();
        stemmer = PorterStemmer.getInstance();
    }

    /**
     * Builds an engine from the maxNofFiles first files in the folder specified.
     * Can only be called once. Writes the engine to disk.
     */
    public void buildIndexFromFolder(String folder, int maxNofFiles) {
        //Get absolute folder (needed by the web-app)
        File f = new File(folder);
        documentFolder = f.getAbsolutePath();

        nofFilesLeft = maxNofFiles;
        nofDocs = nofFilesLeft;
        documents = new Document[nofDocs];
        interval = nofDocs / 10;

        System.out.print("Reading documents: ");
        System.out.print("[");
        StopWatch t = new StopWatch();
        addDocumentsFromFolder(folder);
        System.out.println("]. Took " + t.reset() + "ms.");
        buildIndex();
        System.out.println("Index built in " + t.reset() + "ms. Postings written to " + postingsFilename);
        writeVocabularyToFile();
        System.out.println("Vocabulary written to " + vocabularyFilename + " in " + t.reset() + "ms.");
        writeDocInfoToFile();
        System.out.println("DocInfo written to " + docInfoFilename + " in " + t.reset() + "ms.");
    }

    private void addDocument(File file, int docID) {
        Object[] tokensAndPointers = tokenizeFile(file);
        String[] tokens = (String[]) tokensAndPointers[0];
        Integer[] filepointers = (Integer[]) tokensAndPointers[1];
        if (maxTokensPerDocument != -1 && tokens.length > maxTokensPerDocument) {
            //Restrict length to maxTokensPerDocument (if its not -1)
            String[] tokens2 = new String[maxTokensPerDocument];
            System.arraycopy(tokens, 0, tokens2, 0, maxTokensPerDocument);
            tokens = tokens2;
        }
        int nofTokens = tokens.length;

        //Add occurrences
        for (int i = 0; i < nofTokens; i++) {
            String token = tokens[i];
            if (!termsToFreqList.containsKey(token)) {
                //Create new list to hold freqs for this token
                ArrayList list = new ArrayList(2);
                list.add(new Frequency(docID, 1, filepointers[i].intValue()));
                termsToFreqList.put(token, list);
            } else {
                //Add occurrence for current word
                ArrayList freqs = (ArrayList) termsToFreqList.get(token);
                boolean found = false;
                for (int j = 0; j < freqs.size(); j++) {
                    Frequency freq = (Frequency) freqs.get(j);
                    if (freq.docID == docID) {
                        freq.freq++;
                        found = true;
                        break;
                    }
                }
                if (!found) {
                    //this docID wasn't already in the freqs list
                    freqs.add(new Frequency(docID, 1, filepointers[i].intValue()));
                }
            }
        }
    }

    private void addDocumentsFromFolder(String folderPath) {
        File folder = new File(folderPath);

        //Fix problem with very long filenames
        if (!folder.canRead())
            return;

        File[] files = folder.listFiles();
        for (int i = 0; i < files.length; i++) {
            if (nofFilesLeft == 0)
                break;
            File file = files[i];
            if (!file.isDirectory() && file.canRead()) {
                //Add new document
                nofFilesLeft--;
                if (nofFilesLeft % interval == 0)
                    System.out.print("-");

                int index = nofDocs - nofFilesLeft - 1;
                //Create document to save path. Url, title etc will be generated after indexing
                documents[index] = new Document(index, folder.getName() + "/" + file.getName());
                addDocument(file, index);
            } else {
                //Add docs from sub-folder
                addDocumentsFromFolder(file.getPath());
            }
        }
    }

    private void buildIndex() {
        Set keys = termsToFreqList.keySet();
        nofTermsInIndex = termsToFreqList.size();
        ArrayList dictionaryEntries = new ArrayList(nofTermsInIndex);
        highestFreq = new short[nofDocs];
        vectorLengths = new float[nofDocs];

        try {
            postingsFile = new RandomAccessFile(postingsFilename, "rw");
            postingsFile.setLength(0);
            postingsFile.seek(0); //Start writing engine from start of file...

            for (Iterator iterator = keys.iterator(); iterator.hasNext();) {
                //For each term: create postingslist and write to file
                String word = (String) iterator.next();
                ArrayList freqs = (ArrayList) termsToFreqList.get(word);
                int nofTermDocs = freqs.size();

                //Count freqs for current word (sum over all documents):
                for (int i = 0; i < freqs.size(); i++) {
                    Frequency frequency = (Frequency) freqs.get(i);
                    if (frequency.freq > highestFreq[frequency.docID])
                        highestFreq[frequency.docID] = (short) frequency.freq;
                }

                //Only add dictionaryEntry if it occurs in more than one document or Settings.REMOVE.. isnt set
                if (!Settings.REMOVE_SINGLE_DOC_INDEX_TERMS || nofTermDocs > 1) {
                    dictionaryEntries.add(new VocabularyEntry(word, nofTermDocs, postingsFile.getFilePointer()));

                    //Write all (non-zero) freqs to Posting-file
                    for (int i = 0; i < freqs.size(); i++) {
                        Frequency frequency = (Frequency) freqs.get(i);
                        postingsFile.writeInt(frequency.docID);
                        postingsFile.writeShort(frequency.freq);
                        postingsFile.writeInt(frequency.filePointer);
                    }
                }
            }

            if (Settings.REMOVE_SINGLE_DOC_INDEX_TERMS)
                System.out.println("nofIndexTerms is was reduced from " + nofTermsInIndex + " to " + dictionaryEntries.size());
            else
                System.out.println("nofIndexTerms is " + nofTermsInIndex);

            //Create actual vocabulary (all terms occurring in only one file are removed)
            nofTermsInIndex = dictionaryEntries.size();
            vocabulary = (VocabularyEntry[]) dictionaryEntries.toArray(new VocabularyEntry[nofTermsInIndex]);

            //Sort vocabulary to allow binary searching
            Arrays.sort(vocabulary);

            //Calculate vectorLengths
            for (int i = 0; i < vocabulary.length; i++) {
                VocabularyEntry de = vocabulary[i];
                postingsFile.seek(de.filePointer);
                for (int j = 0; j < de.nofDocs; j++) {
                    int docID = postingsFile.readInt();
                    int freq = postingsFile.readShort();
                    postingsFile.readInt(); //Skip the filePointer
                    float tf = (float) freq / highestFreq[docID];
                    float idf = (float) Math.log((float) nofDocs / de.nofDocs);
                    float tfIdf = tf * idf;
                    vectorLengths[docID] += tfIdf * tfIdf;
                }
            }
            for (int i = 0; i < vectorLengths.length; i++) {
                vectorLengths[i] = (float) Math.sqrt(vectorLengths[i]);
            }

            //Save and close
            postingsFile.close();

        } catch (IOException ioe) {
            ioe.printStackTrace();
        }

        //Release resources
        termsToFreqList = null;
    }

    //Word = any 2 or more consecutive normal letters
    //Returns {String[] tokens and int[] filepointers}
    private Object[] tokenizeFile(File file) {
        try {
            BufferedReader in = new BufferedReader(new FileReader(file));
            StringBuffer sb = new StringBuffer();
            int val = in.read();
            int filepointer = 0; //Block addressing
            int currentPosition = 0;
            ArrayList tokens = new ArrayList(100);
            ArrayList filepointers = new ArrayList(100);
            while (val != -1) {
                if ((val >= 65 && val <= 90) || (val >= 97 && val <= 122)) {
                    //Found part of a word (A-Z || a-z)
                    sb.append((char) val);
                } else {
                    //Found specialchar.
                    //If len>1 make word
                    if (sb.length() > 1) {

                        String word = sb.toString().toLowerCase();
                        if (!Settings.REMOVE_SNIPPET_STOPWORDS || !stopwordRemover.isStopword(word)) {
                            filepointers.add(new Integer(filepointer));
                            if (Settings.USE_STEMMING)
                                word = stemmer.getStemmedWord(word);
                            tokens.add(word);
                        }

                    }
                    //If chars in buffer then reset it
                    if (sb.length() > 0)
                        sb = new StringBuffer();
                }

                //Start new block if filepointer is more than 150 chars from currentPos
                //or a sentence divider is encountered.
                char temp = (char) val;
                if (currentPosition - filepointer > 150 || temp == '.')
                    //Start next block at . + 2, this works because if it's at the end then
                    //no more tokens will be added
                    filepointer = currentPosition + 2;


                val = in.read();
                currentPosition++;
            }

            //Add last token (potentially)
            if (sb.length() > 1) {
                String word = sb.toString().toLowerCase();
                if (!Settings.REMOVE_SNIPPET_STOPWORDS || !stopwordRemover.isStopword(word)) {
                    filepointers.add(new Integer(filepointer - word.length()));
                    if (Settings.USE_STEMMING)
                        word = stemmer.getStemmedWord(word);
                    tokens.add(word);
                }
            }

            String[] tokenz = (String[]) tokens.toArray(new String[tokens.size()]);
            Integer[] filepointerz = (Integer[]) filepointers.toArray(new Integer[filepointers.size()]);
            return new Object[]{tokenz, filepointerz};
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
    }

    private void writeVocabularyToFile() {
        ObjectOutputStream vocabularyFile;
        try {
            //Delete file
            RandomAccessFile vocDel = new RandomAccessFile(vocabularyFilename, "rw");
            vocDel.setLength(0);
            vocDel.close();

            vocabularyFile = new ObjectOutputStream(new FileOutputStream(vocabularyFilename));
            vocabularyFile.writeInt(vocabulary.length);
            //vocabularyFile.writeObject(vocabulary);
            for (int i = 0; i < vocabulary.length; i++) {
                VocabularyEntry dictionaryEntry = vocabulary[i];
                vocabularyFile.writeObject(dictionaryEntry.word);
                vocabularyFile.writeInt(dictionaryEntry.nofDocs);
                vocabularyFile.writeLong(dictionaryEntry.filePointer);
            }
            vocabularyFile.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private void writeDocInfoToFile() {
        ObjectOutputStream docInfoFile;
        try {
            //Delete file
            RandomAccessFile docDel = new RandomAccessFile(docInfoFilename, "rw");
            docDel.setLength(0);
            docDel.close();

            docInfoFile = new ObjectOutputStream(new FileOutputStream(docInfoFilename));
            docInfoFile.writeInt(nofDocs);
            docInfoFile.writeObject(highestFreq);
            docInfoFile.writeObject(vectorLengths);
            docInfoFile.writeObject(documentFolder);

            //Free some resources before building documents
            vocabulary = null;
            highestFreq = null;
            vectorLengths = null;
            System.gc();
            Thread.yield();

            docInfoFile.writeObject(createDocuments());
            docInfoFile.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private Document[] createDocuments() {
        Document[] docs = new Document[nofDocs];
        for (int i = 0; i < nofDocs; i++) {
            Document doc = documents[i]; //Only filename is assigned.

            //Get url
            String url = doc.filename;
            url = "http://" + url.substring(url.lastIndexOf('/') + 1, url.length() - 4);
            url = url.replaceAll("#", "/");
            url = url.replaceAll("@", "?");
            url = url.replaceAll("_", ":");
            url = url.replaceAll("", "_");
            doc.url = url;

            try {
                BufferedReader br = new BufferedReader(new FileReader(documentFolder + "/" + doc.filename));
                String title = br.readLine();
                title = title.substring(0, title.length() - 1); //Strip . at end of title

                //Trunc long titles
                if (title.length() > 70) {
                    title = title.substring(0, 70) + "...";
                }

                br.close();
                doc.title = title;

            } catch (Exception e) {
                System.out.println("Couldn't get snippet and title for " + doc.filename);
            }

            docs[i] = doc;
        }

        return docs;
    }

    /**
     * Convenience class to hold posting triplets before writing them to file
     */
    private class Frequency {

        public final int docID;
        public int freq;
        public final int filePointer;

        public Frequency(int docID, int freq, int filePointer) {
            this.docID = docID;
            this.freq = freq;
            this.filePointer = filePointer;
        }
    }
}
